# -*- coding: utf-8 -*-

from typing import List, Optional, Union

import pandas as pd

from pandas_select import iterutils
from pandas_select.label import LabelSelector


Dtypes = Union[str, List[str], type, List[type]]


class HasDtype(LabelSelector):
    """Select columns based on the column dtypes.

    Parameters
    ----------
    include, exclude: scalar or list-like
        A selection of dtypes or strings to be included/excluded. At least one of
        these parameters must be supplied.

    Raises
    ------
    ValueError
        If both of ``include`` and ``exclude`` are empty;
        if ``include`` and ``exclude`` have overlapping elements;
        if any kind of string dtype is passed in.

    Notes
    -----
    * To select all *numeric* types, use ``numpy.number``, ``'number'``
      or `:class:`AllNumeric`.
    * To select strings you must use the ``object``, ``string`` dtype if pandas
      version > 1.0.0, or or :class:`AllStr`
    * See the `numpy dtype hierarchy
      <http://docs.scipy.org/doc/numpy/reference/arrays.scalars.html>`__
    * To select datetimes, use :class:`numpy.datetime64`, ``'datetime'`` or
      ``'datetime64'``.
    * To select timedeltas, use :class:`numpy.timedelta64`, ``'timedelta'`` or
      ``'timedelta64'``.
    * To select Pandas categorical dtypes, use ``'category'`` or :class:`AllCat`.
    * To select Pandas datetimetz dtypes, use ``'datetimetz'``
      or ``'datetime64[ns, tz]'``.

    See Also
    --------
    :meth:`pandas.DataFrame.select_dtypes`: base implementation

    AllBool, AllNumeric, AllNominal

    Examples
    --------
    >>> df = pd.DataFrame({"a": [1, 2],
    ...                    "b": [True, False],
    ...                    "c": [1.0, 2.0]})
    >>> df
       a      b    c
    0  1   True  1.0
    1  2  False  2.0
    >>> df[HasDtype("int")]
       a
    0  1
    1  2
    >>> import numpy as np
    >>> df[HasDtype(include=np.number, exclude=["int"])]
         c
    0  1.0
    1  2.0
    """

    def __init__(
        self, include: Optional[Dtypes] = None, exclude: Optional[Dtypes] = None,
    ) -> None:
        super().__init__(axis="columns", level=None)
        self.include = include and iterutils.to_set(include)
        self.exclude = exclude and iterutils.to_set(exclude)

    def __call__(self, df: pd.DataFrame) -> pd.Index:
        df_row = df.iloc[:1]
        return df_row.select_dtypes(self.include, self.exclude).columns


class AllNumeric(HasDtype):
    """
    Select numeric columns.

    See Also
    --------
    HasDtype

    Examples
    --------
    >>> df = pd.DataFrame({"a": [1, 2],
    ...                    "b": [True, False],
    ...                    "c": [1.0, 2.0]})
    >>> df
       a      b    c
    0  1   True  1.0
    1  2  False  2.0
    >>> df[AllNumeric()]
       a    c
    0  1  1.0
    1  2  2.0
    """

    def __init__(self) -> None:
        super().__init__("number")


class AllBool(HasDtype):
    """Select boolean columns.

    See Also
    --------
    HasDtype

    Examples
    --------
    >>> df = pd.DataFrame({"a": [1, 2],
    ...                    "b": [True, False],
    ...                    "c": [1.0, 2.0]})
    >>> df
       a      b    c
    0  1   True  1.0
    1  2  False  2.0
    >>> df[AllBool()]
           b
    0   True
    1  False
    """

    def __init__(self) -> None:
        super().__init__("bool")


class AllCat(HasDtype):
    """Select categorical columns.

    Parameters
    ----------
    ordered: default None
       Whether to filter ordered categorical, `None` to select all categorical columns.

    See Also
    --------
    HasDtype, AllNominal

    Examples
    --------
    >>> df = pd.DataFrame({"i": [1, 2],
    ...                    "cat":  pd.Categorical(["a", "b"], ordered=False),
    ...                    "ordered_cat": pd.Categorical(["a", "b"], ordered=True)})
    >>> df
       i cat ordered_cat
    0  1   a           a
    1  2   b           b
    >>> df[AllCat()]
      cat ordered_cat
    0   a           a
    1   b           b
    >>> df[AllCat(ordered=True)]
      ordered_cat
    0           a
    1           b
    """

    def __init__(self, *, ordered: Optional[bool] = None) -> None:
        super().__init__("category")
        self.ordered = ordered

    def __call__(self, df: pd.DataFrame) -> pd.Index:
        cols = super().__call__(df)
        if self.ordered is not None:
            drop_cols = [col for col in cols if df[col].cat.ordered != self.ordered]
            cols = cols.drop(drop_cols)
        return cols


def _get_str_dtypes(strict: bool) -> List[str]:
    old_pandas = pd.__version__ < "1.0.0"  # noqa: WPS609

    if strict:
        if old_pandas:
            raise ValueError("strict=True is incompatible with pandas < 1.0.0")
        return ["string"]

    return ["object"] if old_pandas else ["string", "object"]


class AllStr(HasDtype):
    """Select columns with dtype ``object``, and ``string`` if pandas version >= 1.0.0.

    Parameters
    ----------
    strict: default False
       If True, filter out `object` dtypes.

    Raises
    ------
    ValueError
        If pandas version < 1.0.0 and ``strict = True``

    Notes
    -----
    Be aware that ``strict=False``, the default,  will select **all** `object`
    dtype columns. Columns with mixed types are stored with the `object` dtype!

    See Also
    --------
    HasDtype, AllNominal

    Examples
    --------
    >>> df = pd.DataFrame({"i": [1, 2],
    ...                    "o": ["a", 2],
    ...                    "obj_str": ["a", "b"],
    ...                    "str": ["a", "b"]})
    >>> try:
    ...     df = df.astype({"obj_str": "object", "str": "string"})
    ... except TypeError: # pandas.__version__ < '1.0.0'
    ...     df = df.astype({"obj_str": "object", "str": "object"})
    >>> df
       i  o obj_str str
    0  1  a       a   a
    1  2  2       b   b
    >>> df[AllStr()]
       o obj_str str
    0  a       a   a
    1  2       b   b
    >>> pd.__version__ >= '1.0.0'
    True
    >>> df[AllStr(strict=True)]
      str
    0   a
    1   b
    """

    def __init__(self, *, strict: bool = False):
        self.strict = strict
        super().__init__(_get_str_dtypes(self.strict))


class AllNominal(HasDtype):
    """Select nominal columns.

    Columns with the following dtypes are considered nominal: ``category``, ``object``,
    and ``string`` if pandas version >= 1.0.0.

    See Also
    --------
    HasDtype, AllCat, AllStr

    Examples
    --------
    >>> df = pd.DataFrame({"a": [1, 2],
    ...                    "b": ["a", "b"],
    ...                    "c": ["a", "b"]})
    >>> df = df.astype({"a": "int", "b": "object", "c": "category"})
    >>> df
       a  b  c
    0  1  a  a
    1  2  b  b
    >>> df[AllNominal()]
       b  c
    0  a  a
    1  b  b
    """

    def __init__(self, *, strict: bool = False) -> None:
        self.strict = strict
        dtypes = ["category", *_get_str_dtypes(strict)]
        super().__init__(dtypes)



# -*- coding: utf-8 -*-
import typing

from typing import Any, List, Set, Tuple, Union

import pandas as pd

from pandas.api.types import is_list_like


def to_list(obj: Any) -> List:
    """Wrap the object in a list if it is not already list-like.

    Strings and tuples are not considered list-like.
    """
    if is_list_like(obj) and not isinstance(obj, (str, tuple)):
        return obj
    return [obj]


def to_set(obj: Any) -> Set:
    """Wrap the object in a set if it is not already a set."""
    if isinstance(obj, set):
        return obj
    return set(to_list(obj))


# flake8: noqa: C901, WPS210, WPS221, WPS435, WPS437, WPS507
def mi_intersection(left: pd.MultiIndex, right: pd.MultiIndex) -> pd.MultiIndex:
    """ Intersection of MultiIndexes, preserving order.
    Fix https://github.com/pandas-dev/pandas/issues/31325
    """

    if left.equals(right):
        return left

    lvals = left._ndarray_values
    rvals = right._ndarray_values

    uniq_tuples = None  # flag whether _inner_indexer was successful
    if left.is_monotonic and right.is_monotonic:
        try:
            uniq_tuples = left._inner_indexer(lvals, rvals)[0]
        except TypeError:
            pass  # noqa:WPS420
    if uniq_tuples is None:
        right_uniq = set(rvals)
        seen: Set[Tuple] = set()
        uniq_tuples = [
            x
            for x in lvals
            if x in right_uniq and not (x in seen or seen.add(x))  # type: ignore
        ]
    names = left.names if left.names == right.names else None

    if len(uniq_tuples) == 0:
        return pd.MultiIndex(
            levels=left.levels,
            codes=[[]] * left.nlevels,
            names=names,
            verify_integrity=False,
        )
    return pd.MultiIndex.from_tuples(uniq_tuples, sortorder=0, names=names)



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from typing import Any, Callable, Iterable, Optional, Union

import numpy as np
import pandas as pd

from pandas.api.types import is_bool_dtype, is_list_like

from pandas_select import iterutils
from pandas_select.base import LogicalOp, PrettyPrinter, Selector


Cond = Callable[[pd.Series], Iterable[bool]]
Columns = Union[str, Iterable[str], Callable]


class _BoolIndexerMixin(PrettyPrinter, ABC):
    """Base class for selectors that filters rows by value."""

    def __init__(self, cond: Cond, columns: Optional[Columns] = None):
        self.cond = cond
        if callable(columns):
            self.columns = columns
        else:
            self.columns = columns and iterutils.to_set(columns)  # type:ignore

    def __call__(self, df: pd.DataFrame) -> Iterable[bool]:
        """Apply the condition to each column.

        Return a boolean array with size `df.shape[0]`.
        """
        if self.columns is not None:
            df = df[self.columns]
        masks = df.apply(self.cond)
        return self._join(masks)

    @abstractmethod
    def _join(self, df: pd.DataFrame) -> Iterable[bool]:
        raise NotImplementedError()


class _BoolOpsMixin:
    """Common logical operators mixin."""

    def intersection(self, other: _BoolIndexerMixin) -> "BoolOp":
        """Select elements in both selectors."""
        return BoolOp(np.logical_and, "&", self, other)  # type: ignore

    def union(self, other: _BoolIndexerMixin) -> "BoolOp":
        """Select elements in the left side but not in right side."""
        return BoolOp(np.logical_or, "|", self, other)  # type: ignore

    def symmetric_difference(self, other: Any) -> "BoolOp":
        """Select elements either in the left side or the right side but not in both."""
        return BoolOp(np.logical_xor, "^", self, other)  # type: ignore

    def __and__(self, other: _BoolIndexerMixin) -> "BoolOp":
        return self.intersection(other)

    def __rand__(self, other: _BoolIndexerMixin) -> "BoolOp":
        return BoolOp(np.logical_and, "&", other, self)  # type: ignore

    def __or__(self, other: _BoolIndexerMixin) -> "BoolOp":
        return self.union(other)

    def __ror__(self, other: _BoolIndexerMixin) -> "BoolOp":
        return BoolOp(np.logical_or, "|", other, self)  # type: ignore

    def __xor__(self, other: _BoolIndexerMixin) -> "BoolOp":
        return self.symmetric_difference(other)

    def __rxor__(self, other: _BoolIndexerMixin) -> "BoolOp":
        return BoolOp(np.logical_xor, "^", other, self)  # type: ignore

    def __invert__(self) -> "BoolOp":
        return BoolOp(np.invert, "~", self)  # type: ignore


class _BoolMask(PrettyPrinter, _BoolOpsMixin):
    def __init__(self, mask: Iterable[bool]):
        self.mask = np.asarray(mask)

    def __call__(self, df: pd.DataFrame) -> Iterable[bool]:
        return self.mask


BoolOperation = Callable[[np.ndarray, Optional[np.ndarray]], np.ndarray]
BoolOperand = Union[Selector, Iterable[bool]]


class BoolOp(LogicalOp, _BoolOpsMixin):
    """A logical operation between two `Where` selectors."""

    def __init__(
        self,
        op: BoolOperation,
        op_name: str,
        left: BoolOperand,
        right: Optional[BoolOperand] = None,
    ):
        bool_selectors = [self._validate_operand(operand) for operand in (left, right)]
        super().__init__(op, op_name, *bool_selectors)  # type:ignore

    def _validate_operand(self, sel: Any) -> Union[Selector, Iterable[bool]]:
        if sel is None or callable(sel):
            return sel

        if not is_list_like(sel):
            raise TypeError("Operand does not support logical operations.")

        sel = np.asarray(sel)
        if not is_bool_dtype(sel):
            raise TypeError(f"Operand is not boolean dtype.")

        return _BoolMask(sel)


class BoolIndexer(_BoolIndexerMixin, _BoolOpsMixin, ABC):
    """Base class for boolean indexing and support logical operations."""


class Anywhere(BoolIndexer):
    """Filter rows where *any* column matches a condition.

    Parameters
    ----------
    cond: callable or boolean array-like
        Select labels where `cond` is True. If `cond` is a callable, it is computed on
        each column and should return a boolean array.

    columns: optional
        Subset of columns on which to apply `cond`.

    Examples
    --------
    >>> df = pd.DataFrame({'A': [1, 1, 4], 'B': [2, -3, 1]}, index=["a", "b", "c"])
    >>> df
       A  B
    a  1  2
    b  1 -3
    c  4  1
    >>> df.loc[Anywhere(lambda x : x % 2 == 0)]
       A  B
    a  1  2
    c  4  1
    >>> df.loc[Anywhere(lambda x : x % 2 == 0, columns="A")]
       A  B
    c  4  1
    """

    def _join(self, df: pd.DataFrame) -> np.ndarray:
        return df.any(axis="columns").values


class Everywhere(BoolIndexer):
    """Filter rows where *all* columns match a condition.

    Parameters
    ----------
    cond: callable or boolean array-like
        Select labels where `cond` is True. If `cond` is a callable, it is computed on
        each column and should return a boolean array.

    columns: optional
        Subset of columns on which to apply `cond`.

    Examples
    --------
    >>> df = pd.DataFrame({'A': [1, 1, 4], 'B': [2, -3, 1]}, index=["a", "b", "c"])
    >>> df
       A  B
    a  1  2
    b  1 -3
    c  4  1
    >>> df.loc[Everywhere(lambda x : x > 0)]
       A  B
    a  1  2
    c  4  1
    >>> df.loc[Everywhere(lambda x : x == 1, columns="A")]
       A  B
    a  1  2
    b  1 -3
    """

    def _join(self, df: pd.DataFrame) -> np.ndarray:
        return df.all(axis="columns").values



# -*- coding: utf-8 -*-

from collections import Counter
from functools import partial
from typing import Any, Callable, Iterable, Optional, Union, cast

import numpy as np
import pandas as pd

from pandas.util import Substitution

from pandas_select import iterutils
from pandas_select.base import LogicalOp, PrettyPrinter
from pandas_select.bool import Everywhere


Axis = Union[int, str]

AXIS_DOC = (
    "axis: default 'columns'\n"
    + "\tAxis along which the function is applied, {0 or 'index', 1 or 'columns'}\n"
)

LEVEL_DOC = (
    "level: optional\n"
    + "\tEither the integer position of the level or its name.\n"
    + "\tIt should only be set if ``axis`` targets a MultiIndex, "
    + "otherwise a :exc:`IndexError` will be raised.\n"
)


def _validate_axis(axis: Axis) -> Axis:
    allowed = [0, 1, "index", "columns"]
    if axis not in allowed:
        raise ValueError(f"axis must be one of {allowed}.")
    return axis


def _validate_indexer(indexer: Iterable) -> Iterable:
    """Ensure `indexer` can be used as an indexer on another index.
    https://pandas.pydata.org/pandas-docs/stable/user_guide/missing_data.html
    """
    if not isinstance(indexer, pd.MultiIndex) and pd.isna(indexer).any():
        # isna is not defined for MultiIndex
        return list(indexer)
    return indexer


@Substitution(axis=AXIS_DOC, level=LEVEL_DOC)
class _LabelSelectorMixin(PrettyPrinter):
    """
    Base class for selecting indexes or columns based on their labels.

    Parameters
    ----------
    %(axis)s
    %(level)s
    """

    def __init__(self, axis: Axis = "columns", level: Optional[Axis] = None):
        self.axis = _validate_axis(axis)
        self.level = level

    def __call__(self, df: pd.DataFrame) -> Iterable:
        labels = df._get_axis(self.axis)  # noqa: WPS437
        if self.level is not None:
            index = labels.get_level_values(self.level)
        else:
            index = labels

        selection: pd.Index = labels[self._get_indexer(index)]

        if selection.has_duplicates:
            raise RuntimeError(f"Found duplicated values in selection")

        return _validate_indexer(selection)

    def _get_indexer(self, index: pd.Index) -> Iterable:
        raise NotImplementedError()


def _intersection(left: pd.Index, right: pd.Index) -> pd.Index:
    if (  # noqa: WPS337
        isinstance(left, pd.MultiIndex)
        and isinstance(right, pd.MultiIndex)
        and pd.__version__ < "1.1.0"  # noqa: WPS609
    ):
        return iterutils.mi_intersection(left, right)
    return left.intersection(right, sort=False)


def _union(left: pd.Index, right: pd.Index) -> pd.Index:
    return left.union(right, sort=False)


def _difference(left: pd.Index, right: pd.Index) -> pd.Index:
    return left.difference(right, sort=False)


def _symmetric_difference(left: pd.Index, right: pd.Index) -> pd.Index:
    return left.symmetric_difference(right, sort=False)


class _LabelOpsMixin:
    """Common logical operators mixin."""

    def intersection(self, other: Any) -> "LabelOp":
        """Select elements in both selectors."""
        return LabelOp(_intersection, "&", self, other)  # type:ignore

    def union(self, other: Any) -> "LabelOp":
        """Select elements in the left side but not in the right side."""
        return LabelOp(_union, "|", self, other)  # type:ignore

    def difference(self, other: Any) -> "LabelOp":
        """Select elements in the left side but not in the right side."""
        return LabelOp(_difference, "-", self, other)  # type:ignore

    def symmetric_difference(self, other: Any) -> "LabelOp":
        """Select elements that are either in the left side or the right side
        but not in both.
        """
        return LabelOp(
            _symmetric_difference, "^", self, other  # type:ignore
        )

    def __and__(self, other: Any) -> "LabelOp":
        return self.intersection(other)

    def __rand__(self, other: Any) -> "LabelOp":
        return LabelOp(_intersection, "&", other, self)  # type:ignore

    def __or__(self, other: Any) -> "LabelOp":
        return self.union(other)

    def __ror__(self, other: Any) -> "LabelOp":
        return LabelOp(_union, "|", other, self)  # type:ignore

    def __sub__(self, other: Any) -> "LabelOp":
        return self.difference(other)

    def __rsub__(self, other: Any) -> "LabelOp":
        return LabelOp(_difference, "-", other, self)  # type:ignore

    def __xor__(self, other: Any) -> "LabelOp":
        return self.symmetric_difference(other)

    def __rxor__(self, other: Any) -> "LabelOp":
        return LabelOp(_symmetric_difference, "^", other, self)  # type:ignore

    def __invert__(self) -> "LabelOp":
        return LabelInvertOp(self)  # type:ignore


def _to_index(obj: Union[Iterable, pd.Index]) -> pd.Index:
    if isinstance(obj, pd.Index):
        return obj

    obj = iterutils.to_list(obj)
    if isinstance(obj[0], tuple):
        return pd.MultiIndex.from_tuples(obj)

    return pd.Index(obj)


class LabelOp(LogicalOp, _LabelOpsMixin):
    """A logical operation between two `:class:`_LabelSelectorMixin` selectors."""

    def __init__(
        self,
        op: Callable[[Iterable, Optional[Iterable]], Iterable],
        op_name: str,
        left: _LabelSelectorMixin,
        right: Optional[_LabelSelectorMixin] = None,
    ):
        left = self._validate_selector(left)

        if right is not None:
            right = self._validate_selector(
                right, getattr(left, "axis", "columns"), getattr(left, "level", None)
            )

            if left.axis != right.axis:
                raise ValueError(f"{left} and {right} must target the same axis.")

        super().__init__(op, op_name, left, right)
        self.axis = left.axis

    def __call__(self, df: pd.DataFrame) -> Iterable:
        lvals = _to_index(self.left(df))
        operands = [lvals]

        if self.right is not None:
            rvals = _to_index(self.right(df))
            operands.append(rvals)

        selection = self.op(*operands)
        return _validate_indexer(selection)

    def _validate_selector(
        self, obj: Any, axis: Axis = "columns", level: Optional[int] = None
    ) -> _LabelSelectorMixin:
        if callable(obj):
            return cast(_LabelSelectorMixin, obj)
        return Exact(obj, axis=axis, level=level)


class LabelSelector(_LabelSelectorMixin, _LabelOpsMixin):
    """Base class for label selection and support logical operations."""


class LabelInvertOp(LabelOp):
    """Invert operation on a :class:`_LabelSelectorMixin`."""

    def __init__(self, selector: LabelSelector):
        super().__init__(np.logical_not, "~", selector)
        self.axis = selector.axis
        self.level = selector.level

    def __call__(self, df: pd.DataFrame) -> pd.Index:
        index = df._get_axis(self.axis)  # noqa: WPS437
        if self.level is not None:
            level_index = index.get_level_values(self.level)
        else:
            level_index = index

        values = self.left(df).get_level_values(self.level)  # type: ignore
        selection = index[~level_index.isin(values)]

        return _validate_indexer(selection)


@Substitution(axis=AXIS_DOC, level=LEVEL_DOC)
class Exact(LabelSelector):
    """
    Select labels from a list,
    sorted by the order they appear in the list.

    Parameters
    ----------
    values: single label or list-like
        Index or column labels to select

    %(axis)s
    %(level)s

    Raises
    ------
    ValueError:
        If ``values`` contains duplicates.

    Examples
    --------
    >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["a", "b"])
    >>> df
       A  B
    a  1  3
    b  2  4
    # Same as df[["B", "A"]]:
    >>> df[Exact(["B", "A"])]
       B  A
    a  3  1
    b  4  2
    # Same as df.loc[["b"]]:
    >>> df.loc[Exact("b", axis="index")]
       A  B
    b  2  4
    """

    def __init__(
        self,
        values: Union[Any, Iterable],
        axis: Axis = "columns",
        level: Optional[int] = None,
    ):
        super().__init__(axis, level)
        self.values = self._validate_values(values)

    def _validate_values(self, values: Union[Any, Iterable]) -> Iterable:
        values = iterutils.to_list(values)
        counts = Counter(values)
        dups = [val for val, cnt in counts.items() if cnt > 1]
        if dups:
            raise ValueError(f"Found duplicated values")
        return values

    def _get_indexer(self, index: pd.Index) -> Union[Iterable[int], np.ndarray]:
        indexer = index.get_indexer_for(self.values)

        missing_mask = indexer == -1
        if missing_mask.any():
            missing = np.asarray(self.values)[missing_mask].tolist()
            raise KeyError(missing)

        return indexer


@Substitution(axis=AXIS_DOC, level=LEVEL_DOC)
class AnyOf(LabelSelector):
    """Select labels from a list.

    The labels are sorted by the order they appear in the :class:`~pandas.DataFrame`.

    ``AnyOf`` is similar to :meth:`pandas.Series.isin`.

    Parameters
    ----------
    values: single label or list-like
        Index or column labels to select

    %(axis)s
    %(level)s

    Notes
    -----
    ``AnyOf`` is similar to :meth:`pandas.Series.isin`.

    Examples
    --------
    >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["a", "b"])
    >>> df
       A  B
    a  1  3
    b  2  4
    >>> df[AnyOf(["B", "A", "invalid"])]
       A  B
    a  1  3
    b  2  4
    >>> df.loc[AnyOf("b", axis="index")]
       A  B
    b  2  4
    """

    def __init__(
        self, values: Any, axis: Axis = "columns", level: Optional[int] = None,
    ):
        super().__init__(axis, level)
        self.values = iterutils.to_set(values)  # noqa: WPS110

    def _get_indexer(self, index: pd.Index) -> np.ndarray:
        return index.isin(self.values)


@Substitution(axis=AXIS_DOC, level=LEVEL_DOC)
class AllOf(AnyOf):
    """Same as :class:`AnyOf`, except that a :exc:`KeyError` is raised for labels
    that don't exist.

    Parameters
    ----------
    values: single label or list-like
        Index or column labels to select

    %(axis)s
    %(level)s

    Examples
    --------
    >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["a", "b"])
    >>> df
       A  B
    a  1  3
    b  2  4
    >>> df[AllOf(["B", "A"])]
       A  B
    a  1  3
    b  2  4
    >>> df[AllOf(["B", "A", "invalid"])]
    Traceback (most recent call last):
    ...
    KeyError: {'invalid'}
    """

    def __call__(self, df: pd.DataFrame) -> pd.Index:
        selected = super().__call__(df)

        missing = self.values.difference(selected)
        if missing:
            raise KeyError(missing)

        return selected

    def __invert__(self) -> LabelInvertOp:
        return LabelInvertOp(AnyOf(self.values))


@Substitution(axis=AXIS_DOC, level=LEVEL_DOC)
class Everything(LabelSelector):
    """
    Select all labels.

    Parameters
    ----------
    %(axis)s
    %(level)s
    """

    def _get_indexer(self, index: pd.Index) -> np.ndarray:
        return np.arange(0, index.size)


MASK_MI_DOC = (
    "If applied to a :class:`~pandas.MultiIndex` with ``level=None``,"
    + " all the levels will be tested."
)

IndexMask = Callable[[pd.Index], Iterable[bool]]


@Substitution(axis=AXIS_DOC, level=LEVEL_DOC, mask_mi=MASK_MI_DOC)
class LabelMask(LabelSelector):
    """
    Select labels where the condition is True.

    Parameters
    ----------
    cond: bool Series/DataFrame, array-like, or callable
        Select labels where cond is True. If `cond` is a callable, it is computed on the
        :class:`~pandas.Index` and should return a boolean array.

    %(axis)s
    %(level)s
    kwargs:
        If ``cond`` is a :func:`callable`, keyword arguments to pass to it.

    Notes
    -----
    %(mask_mi)s

    Examples
    --------
    >>> df = pd.DataFrame({"A": [1], "B": [1]})
    >>> df
       A  B
    0  1  1
    >>> df[LabelMask([True, False])]
       A
    0  1
    >>> df[LabelMask(lambda x: x == "A")]
       A
    0  1
    """

    def __init__(
        self,
        cond: Union[Iterable[bool], IndexMask],
        axis: Axis = "columns",
        level: Optional[int] = None,
        **kwargs: Any,
    ):
        super().__init__(axis, level)
        self.cond = cond
        self.kwargs = kwargs

    def _get_indexer(self, index: pd.Index) -> Iterable[bool]:
        if not callable(self.cond):
            return self.cond

        func = partial(self.cond, **self.kwargs)

        if isinstance(index, pd.MultiIndex):
            mi_df = index.to_frame()
            selected = mi_df[Everywhere(func)]
            mi_values = pd.MultiIndex.from_frame(selected).to_numpy()
            return index.get_locs(mi_values)

        return func(index)


class _IgnoreCase(LabelMask):
    def __init__(  # noqa: WPS211
        self,
        cond: Callable[[np.ndarray, str], np.ndarray],
        pat: str,
        case: bool = True,
        axis: Axis = "columns",
        level: Optional[int] = None,
        **kwargs: Any,
    ):
        kwargs["pat"] = pat if case else pat.lower()
        super().__init__(cond, axis, level, **kwargs)  # type:ignore
        self.case = case

    def _get_indexer(self, index: pd.Index) -> np.ndarray:
        index = index if self.case else index.str.lower()
        return super()._get_indexer(index)


PAT_DOC = "pat:\n\tCharacter sequence. Regular expressions are not accepted."
CASE_DOC = "case: default True\n\tIf True, case sensitive."


@Substitution(
    axis=AXIS_DOC, level=LEVEL_DOC, pat=PAT_DOC, case=CASE_DOC, mask_mi=MASK_MI_DOC
)
class StartsWith(_IgnoreCase):
    """
    Select labels that start with a prefix.

    Parameters
    ----------
    %(pat)s
    %(case)s
    %(axis)s
    %(level)s

    See Also
    --------
    EndsWith: Same as `StartsWith`, but tests the end of string.

    Notes
    -----
    %(mask_mi)s

    Examples
    --------
    >>> df = pd.DataFrame({"bat": [1], "Bear": [1], "cat": [1]})
    >>> df
       bat  Bear  cat
    0    1     1    1
    >>> df[StartsWith("b")]
       bat
    0    1
    >>> df[StartsWith("b", case=False)]
       bat  Bear
    0    1     1
    """

    def __init__(
        self,
        pat: str,
        case: bool = True,
        axis: Axis = "columns",
        level: Optional[int] = None,
    ):
        super().__init__(
            pd.core.strings.str_startswith, pat, case, axis, level, na=False
        )


@Substitution(
    axis=AXIS_DOC, level=LEVEL_DOC, pat=PAT_DOC, case=CASE_DOC, mask_mi=MASK_MI_DOC
)
class EndsWith(_IgnoreCase):
    """Select labels that end with a suffix.

    Parameters
    ----------
    %(pat)s
    %(case)s
    %(axis)s
    %(level)s

    See Also
    --------
    StartsWith: Same as `EndsWith`, but tests the start of string.

    Notes
    -----
    %(mask_mi)s

    Examples
    --------
    >>> df = pd.DataFrame({"bat": [1], "Bear": [1], "caT": [1]})
    >>> df
       bat  Bear  caT
    0    1     1    1
    >>> df[EndsWith("t")]
       bat
    0    1
    >>> df[EndsWith("t", case=False)]
       bat  caT
    0    1    1
    """

    def __init__(
        self,
        pat: str,
        case: bool = True,
        axis: Axis = "columns",
        level: Optional[int] = None,
    ):
        super().__init__(pd.core.strings.str_endswith, pat, case, axis, level, na=False)


FLAGS_DOC = (
    "flags: default 0, i.e no flags\n\tFlags to pass through to the re module, "
    + "e.g. :data:`re.IGNORECASE`."
)


@Substitution(
    axis=AXIS_DOC, level=LEVEL_DOC, pat=PAT_DOC, case=CASE_DOC, flags=FLAGS_DOC
)
class Contains(LabelMask):
    """Select labels that contain a pattern or regular expression.

    Parameters
    ----------
    %(pat)s
    %(case)s
    %(flags)s
    regex : default True
        If True, assumes that ``pat`` is a regular expression.
        If False, treats that ``pat`` as a literal string.
    %(axis)s
    %(level)s

    See Also
    --------
    :meth:`pandas.Series.str.contains`: Base implementation
    Match: Analogous, but stricter, relying on :func:`re.match` instead of
        :func:`re.search`.

    Examples
    --------
    >>> df = pd.DataFrame({"Mouse": [1], "dog": [1], "house and parrot": [1]})
    >>> df
       Mouse  dog  house and parrot
    0      1    1                 1
    >>> df[Contains("og", regex=False)]
       dog
    0    1
    >>> df[Contains("house|dog")]
       dog  house and parrot
    0    1                 1
    """

    def __init__(  # noqa: WPS211
        self,
        pat: str,
        case: bool = True,
        flags: int = 0,
        regex: bool = True,
        axis: Axis = "columns",
        level: Optional[int] = None,
    ):
        contains_kw = {
            "pat": pat,
            "case": case,
            "flags": flags,
            "na": False,
            "regex": regex,
        }
        super().__init__(pd.core.strings.str_contains, axis, level, **contains_kw)


@Substitution(
    axis=AXIS_DOC, level=LEVEL_DOC, pat=PAT_DOC, case=CASE_DOC, flags=FLAGS_DOC
)
class Match(LabelMask):
    """
    Select labels that match a regular expression.

    Parameters
    ----------
    %(pat)s
    %(case)s
    %(flags)s
    %(axis)s
    %(level)s

    See Also
    --------
    :meth:`pandas.Series.str.match`: Base implementation
    Contains: Analogous, but less strict, relying on :func:`re.search`
        instead of :func:`re.match`.

    Examples
    --------"
    >>> df = pd.DataFrame({"Mouse": [1], "dog": [1], "house and parrot": [1]})
    >>> df
       Mouse  dog  house and parrot
    0      1    1                 1
    >>> df[Match(".*og")]
       dog
    0    1
    >>> df[Match("house|dog")]
       dog  house and parrot
    0    1                 1
    """

    def __init__(
        self,
        pat: str,
        flags: int = 0,
        axis: Axis = "columns",
        level: Optional[int] = None,
    ):
        match_kw = {"pat": pat, "flags": flags, "na": False}
        super().__init__(pd.core.strings.str_match, axis, level, **match_kw)



# -*- coding: utf-8 -*-

from inspect import signature
from typing import Any, Callable, Iterable, List, Optional

import pandas as pd


Selector = Callable[[pd.DataFrame], Iterable]


class PrettyPrinter:
    """Base class for all estimators in pandas-select.

    Notes
    -----
    All selectors should specify all the parameters that can be set at the class level
    in their ``__init__`` as explicit keyword arguments (no ``*args`` or ``**kwargs``).
    """

    def __repr__(self) -> str:
        params = signature(self.__class__).parameters
        args = [
            f"{param}={repr(self.__dict__[param])}"
            for param in params
            if param in self.__dict__  # param is a class attribute
        ]
        return self._format(args)

    def __str__(self) -> str:
        """Same as repr but more concise."""
        params = signature(self.__class__).parameters.items()
        args = []
        for name, param in params:
            if name not in self.__dict__:
                continue  # param is not a class attribute
            value = self.__dict__[name]
            if param.default is param.empty:
                args.append(str(value))
            elif param.default != value:
                args.append(f"{name}={str(value)}")

        return self._format(args)

    def _format(self, args: Optional[List[str]] = None) -> str:
        pretty_args = ", ".join(args) if args else ""
        return f"{type(self).__name__}({pretty_args})"


class LogicalOp:
    """Base class for logical operations in pandas-select."""

    def __init__(
        self,
        op: Callable[[Iterable, Optional[Iterable]], Iterable],
        op_name: str,
        left: Selector,
        right: Optional[Selector] = None,
    ):
        self.op = op
        self.op_name = op_name
        self.left = left
        self.right = right

    def __call__(self, df: pd.DataFrame) -> Iterable:
        operands = [self.left(df)]
        if self.right is not None:
            operands.append(self.right(df))
        return self.op(*operands)

    def __repr__(self) -> str:
        return self._pretty_format(repr)

    def __str__(self) -> str:
        return self._pretty_format(str)

    def _pretty_format(self, fmt: Callable[[Any], str]) -> str:
        pretty_left = fmt(self.left)
        if self.right is None:
            return self.op_name + pretty_left
        pretty_right = fmt(self.right)
        return f"{pretty_left} {self.op_name} {pretty_right}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pandas_select']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=0.25.1']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=1.5.0,<2.0.0'],
 'docs': ['sphinx>=2.4.1,<3.0.0',
          'sphinxcontrib-contentui>=0.2.4,<0.3.0',
          'ipython>=7.12.0,<8.0.0']}

setup_kwargs = {
    'name': 'pandas-select',
    'version': '0.1.5',
    'description': 'Supercharged DataFrame indexing',
    'long_description': '==================================================\n``pandas-select``: Supercharged DataFrame indexing\n==================================================\n\n.. image:: https://readthedocs.org/projects/project-template-python/badge/?version=latest\n   :target: https://pandas-select-python.readthedocs.io/\n   :alt: Documentation status\n\n.. image:: https://img.shields.io/pypi/v/pandas-select.svg\n   :target: https://pypi.org/project/pandas-select/\n   :alt: Latest PyPI version\n\n.. image:: https://img.shields.io/pypi/pyversions/pandas-select.svg\n   :target: https://pypi.org/project/pandas-select/\n   :alt: Python versions supported\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Code style: black\n\n.. image:: https://img.shields.io/badge/style-wemake-000000.svg\n   :target: https://github.com/wemake-services/wemake-python-styleguide\n\n``pandas-select`` is a collection of DataFrame selectors that facilitates indexing\nand selecting data, fully compatible with pandas vanilla indexing.\n\nThe selector functions can choose variables based on their name, data type, arbitrary\nconditions, or any combination of these.\n\n``pandas-select`` is inspired by two R libraries: `tidyselect <https://tidyselect.r-lib.org/reference/select_helpers.html>`_\nand `recipe <https://tidymodels.github.io/recipes/reference/selections.html>`_.\n\nInstallation\n------------\n\n``pandas-select`` is a Python-only package `hosted on PyPI <https://pypi.org/project/pandas-select/>`_.\nThe recommended installation method is `pip <https://pip.pypa.io/en/stable/>`_-installing\ninto a `virtualenv <https://hynek.me/articles/virtualenv-lives/>`_:\n\n.. code-block:: console\n\n   $ pip install pandas-select\n\n\nDesign goals\n------------\n\n.. why-begin\n\n* Fully compatible with `pandas.DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html>`_\n  ``[]`` and `pandas.DataFrame.loc <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.loc.html?highlight=loc#pandas.DataFrame.loc>`_\n  accessors.\n\n* Emphasise readability and conciseness by cutting boilerplate:\n\n.. code-block:: python\n\n    # pandas-select\n    df[AllNumeric()]\n    df[StartsWith("Type") | "Legendary"]\n\n    # vanilla\n    cols = df.select_dtypes(exclude="number").columns\n    df[cols]\n    cond = lambda col : col.startswith("Type") or col == "Legendary"\n    cols = [col for col in df.columns if cond(col)]\n    df[cols]\n\n* Ease the challenges of `indexing with hierarchical index <https://pandas.pydata.org/pandas-docs/stable/user_guide/advanced.html#advanced-indexing-with-hierarchical-index>`_\n  and offers an alternative to `slicers <https://pandas.pydata.org/pandas-docs/stable/user_guide/advanced.html#advanced-mi-slicers>`_\n  when the labels cannot be listed manually.\n\n.. code-block:: python\n\n    # pandas-select\n    name = Contains("Jeff", axis="index", level="Name")\n    df_mi.loc[selector]\n\n    # vanilla\n    selector = df_mi.index.get_level_values("Name").str.contains("Jeff")\n    df_mi.loc[selector]\n\n* Allow *deferred selection* when the DataFrame\'s columns are not known in advance,\n  for example in automated machine learning applications. ``pandas_select`` offers\n  integration with `sklearn <https://scikit-learn.org/stable/modules/generated/sklearn.compose.`make_column_selector.html>`_.\n\n.. code-block:: python\n\n    from pandas_select import AnyOf, AllBool, AllNominal, AllNumeric, ColumnSelector\n    from sklearn.compose import make_column_transformer\n    from sklearn.preprocessing import OneHotEncoder, StandardScaler\n\n    ct = make_column_transformer(\n        (StandardScaler(), ColumnSelector(AllNumeric() & ~AnyOf("Generation"))),\n        (OneHotEncoder(), ColumnSelector(AllNominal() | AllBool() | "Generation"))\n    )\n    ct.fit_transform(df)\n',
    'author': 'Jean-Francois Zinque',
    'author_email': 'jzinque@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jeffzi/pandas-select/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-

from pandas_select.__version__ import __version__
from pandas_select.bool import Anywhere, Everywhere
from pandas_select.column import (
    AllBool,
    AllCat,
    AllNominal,
    AllNumeric,
    AllStr,
    HasDtype,
)
from pandas_select.label import (
    AllOf,
    AnyOf,
    Contains,
    EndsWith,
    Everything,
    Exact,
    LabelMask,
    Match,
    StartsWith,
)
from pandas_select.sklearn import ColumnSelector



# -*- coding: utf-8 -*-

from typing import List

import pandas as pd

from pandas_select.base import PrettyPrinter
from pandas_select.label import LabelSelector


class ColumnSelector(PrettyPrinter):
    """Create a callable compatible with :class:`sklearn.compose.ColumnTransformer`.

    Parameters
    ----------
    selector:
        A label selector, i.e. a :func:`callable` that returns a list of strings.

    Raises
    ------
    ValueError:
        If `selector` is not a callable or doesn't target the "columns" axis.

    Examples
    --------
    >>> from pandas_select import AnyOf, AllBool, AllNominal, AllNumeric, ColumnSelector
    >>> from sklearn.compose import make_column_transformer
    >>> from sklearn.preprocessing import OneHotEncoder, StandardScaler
    >>> make_column_transformer(
    ...    (StandardScaler(), ColumnSelector(AllNumeric() & ~AnyOf("Generation"))),
    ...    (OneHotEncoder(), ColumnSelector(AllNominal() | AllBool() | "Generation"))
    ... )
    """

    def __init__(self, selector: LabelSelector):
        self.selector = selector

        if not callable(selector):
            raise ValueError(f"{selector} is not a callable.")

        try:
            if selector.axis not in {1, "columns"}:
                raise ValueError(
                    f"Cannot make a ColumnSelector from {selector}"
                    + ", which does not target the column axis."
                )
        except AttributeError:
            pass  # noqa: WPS420

    def __call__(self, df: pd.DataFrame) -> List[str]:
        if not isinstance(df, pd.DataFrame):
            raise ValueError("ColumnSelector can only be applied to a DataFrame.")
        cols = self.selector(df)
        try:
            # LabelSelector may return a pandas.Index
            return cols.tolist()  # type: ignore
        except AttributeError:
            return list(cols)
